# coding=utf-8
__author__ = "Dimitris Karkalousos"

# Taken and adapted from:
# https://github.com/khammernik/sigmanet/blob/master/reconstruction/common/mytorch/models/datalayer.py

from typing import Any, List, Optional, Tuple

import torch

from atommic.collections.common.parts.fft import fft2, ifft2
from atommic.collections.common.parts.utils import complex_abs, complex_conj, complex_mul


class ConjugateGradient(torch.autograd.Function):
    """Conjugate Gradient solver for the prox of the data term."""

    @staticmethod
    def complexDot(data1: torch.Tensor, data2: torch.Tensor) -> torch.Tensor:
        """Complex dot product of two tensors."""
        nBatch = data1.shape[0]
        mult = complex_mul(data1, complex_conj(data2))
        re, im = torch.unbind(mult, dim=-1)
        return torch.stack([torch.sum(re.view(nBatch, -1), dim=-1), torch.sum(im.view(nBatch, -1), dim=-1)], -1)

    @staticmethod
    def solve(x0: torch.Tensor, M: torch.Tensor, tol: float, max_iter: int) -> torch.Tensor:
        """Solve the linear system Mx=b using conjugate gradient.

        Parameters
        ----------
        x0 : torch.Tensor
            Initial guess. Shape [batch_size, height, width, 2].
        M : torch.Tensor
            Linear operator. Shape [batch_size, height, width, 2].
        tol : float
            Tolerance for the stopping criterion.
        max_iter : int
            Maximum number of iterations.
        """
        nBatch = x0.shape[0]
        x = torch.zeros(x0.shape).to(x0.device)
        r = x0.clone()
        p = x0.clone()
        x0x0 = (x0.pow(2)).view(nBatch, -1).sum(-1)
        rr = torch.stack([(r.pow(2)).view(nBatch, -1).sum(-1), torch.zeros(nBatch).to(x0.device)], dim=-1)

        it = 0
        while torch.min(rr[..., 0] / x0x0) > tol and it < max_iter:
            it = it + 1
            q = M(p)

            data1 = rr
            data2 = ConjugateGradient.complexDot(p, q)

            re1, im1 = torch.unbind(data1, -1)
            re2, im2 = torch.unbind(data2, -1)
            alpha = torch.stack([re1 * re2 + im1 * im2, im1 * re2 - re1 * im2], -1) / complex_abs(data2) ** 2

            x = x + complex_mul(alpha.reshape(nBatch, 1, 1, 1, -1), p.clone())
            r = r - complex_mul(alpha.reshape(nBatch, 1, 1, 1, -1), q.clone())
            rr_new = torch.stack([(r.pow(2)).view(nBatch, -1).sum(-1), torch.zeros(nBatch).to(x0.device)], dim=-1)
            beta = torch.stack([rr_new[..., 0] / rr[..., 0], torch.zeros(nBatch).to(x0.device)], dim=-1)
            p = r.clone() + complex_mul(beta.reshape(nBatch, 1, 1, 1, -1), p)
            rr = rr_new.clone()
        return x

    # pylint: disable=arguments-differ
    @staticmethod
    def forward(
        ctx: torch.autograd.function,
        z: torch.Tensor,
        _lambda: torch.Tensor,
        y: torch.Tensor,
        sensitivity_maps: torch.Tensor,
        mask: torch.Tensor,
        tol: float,
        max_iter: int,
        fft_centered: bool,
        fft_normalization: str,
        spatial_dims: List[int],
    ) -> torch.Tensor:
        """Forward pass of :class:`ConjugateGradient`.

        Parameters
        ----------
        ctx : torch.autograd.function
            Context object.
        z : torch.Tensor
            Input image. Shape [batch_size, height, width, 2].
        _lambda : torch.Tensor
            Regularization parameter.
        y : torch.Tensor
            Subsampled k-space data. Shape [batch_size, num_coils, height, width, 2].
        sensitivity_maps : torch.Tensor
            Coil sensitivity maps. Shape [batch_size, num_coils, height, width, 2].
        mask : torch.Tensor
            Subsampling mask. Shape [batch_size, 1, height, width, 1].
        tol : float
            Tolerance for the stopping criterion.
        max_iter : int
            Maximum number of iterations.
        fft_centered : bool
            Whether to center the FFT.
        fft_normalization : str
            FFT normalization.
        spatial_dims : List[int]
            Spatial dimensions.

        Returns
        -------
        torch.Tensor
            Output image. Shape [batch_size, height, width, 2].
        """
        ctx.tol = tol
        ctx.max_iter = max_iter
        ctx.fft_centered = fft_centered
        ctx.fft_normalization = fft_normalization
        ctx.spatial_dims = spatial_dims

        def A(x):
            """Adjoint of the forward operator."""
            x = (
                fft2(
                    complex_mul(x.expand_as(sensitivity_maps), sensitivity_maps),
                    centered=fft_centered,
                    normalization=fft_normalization,
                    spatial_dims=spatial_dims,
                )
                * mask
            )
            return torch.sum(x, dim=-4, keepdim=True)

        def AT(x):
            """AT = A^H"""
            return torch.sum(
                complex_mul(
                    ifft2(x * mask, centered=fft_centered, normalization=fft_normalization, spatial_dims=spatial_dims),
                    complex_conj(sensitivity_maps),
                ),
                dim=(-5),
            )

        def M(p):
            r"""M = A^H A + lambda I"""
            return _lambda * AT(A(p)) + p

        x0 = _lambda * AT(y) + z
        ctx.save_for_backward(AT(y), x0, sensitivity_maps, mask, _lambda)

        return ConjugateGradient.solve(x0, M, ctx.tol, ctx.max_iter)

    # pylint: disable=arguments-differ
    @staticmethod
    def backward(
        ctx: torch.autograd.function, grad_x: torch.Tensor
    ) -> tuple[torch.Tensor, Any, None, None, None, None, None, None, None, None]:
        """Backward pass of the conjugate gradient solver.

        Parameters
        ----------
        ctx : torch.autograd.function
            Context object.
        grad_x : torch.Tensor
            Gradient of the output image. Shape [batch_size, height, width, 2].

        Returns
        -------
        Tuple[torch.Tensor, ...]
            Gradient of the input image, regularization parameter, ...
        """
        ATy, rhs, sensitivity_maps, mask, _lambda = ctx.saved_tensors

        def A(x):
            """Adjoint of the forward operator."""
            x = (
                fft2(
                    complex_mul(x.expand_as(sensitivity_maps), sensitivity_maps),
                    centered=ctx.fft_centered,
                    normalization=ctx.fft_normalization,
                    spatial_dims=ctx.spatial_dims,
                )
                * mask
            )
            return torch.sum(x, dim=-4, keepdim=True)

        def AT(x):
            """AT = A^H"""
            return torch.sum(
                complex_mul(
                    ifft2(
                        x * mask,
                        centered=ctx.fft_centered,
                        normalization=ctx.fft_normalization,
                        spatial_dims=ctx.spatial_dims,
                    ),
                    complex_conj(sensitivity_maps),
                ),
                dim=(-5),
            )

        def M(p):
            r"""M = A^H A + lambda I"""
            return _lambda * AT(A(p)) + p

        Qe = ConjugateGradient.solve(grad_x, M, ctx.tol, ctx.max_iter)
        QQe = ConjugateGradient.solve(Qe, M, ctx.tol, ctx.max_iter)

        grad_z = Qe

        grad_lambda = (
            complex_mul(
                ifft2(
                    Qe, centered=ctx.fft_centered, normalization=ctx.fft_normalization, spatial_dims=ctx.spatial_dims
                ),
                complex_conj(ATy),
            ).sum()
            - complex_mul(
                ifft2(
                    QQe, centered=ctx.fft_centered, normalization=ctx.fft_normalization, spatial_dims=ctx.spatial_dims
                ),
                complex_conj(rhs),
            ).sum()
        )

        return grad_z, grad_lambda, None, None, None, None, None, None, None, None


class DataIDLayer(torch.nn.Module):
    """Placeholder for the identity data layer."""

    def __init__(self, *args, **kwargs):
        """Inits :class:`DataIDLayer`."""
        super().__init__()


class DataGDLayer(torch.nn.Module):
    """DataLayer computing the gradient on the L2 data term."""

    def __init__(
        self,
        lambda_init: float,
        learnable: bool = True,
        fft_centered: bool = False,
        fft_normalization: str = "backward",
        spatial_dims: Optional[Tuple[int, int]] = None,
    ):
        """Inits :class:`DataGDLayer`.

        Parameters
        ----------
        lambda_init : float
            Initial value of data term weight lambda.
        learnable : bool
            If True, the data term weight lambda is learnable. Default is ``True``.
        fft_centered : bool
            If True, the FFT is centered. Default is ``False``.
        fft_normalization : str
            If "ortho", the FFT is normalized. Default is ``"backward"``.
        spatial_dims : tuple of int
            If not None, the spatial dimensions of the FFT. Default is ``None``.
        """
        super().__init__()
        self.lambda_init = lambda_init
        self.data_weight = torch.nn.Parameter(torch.Tensor(1))
        self.data_weight.data = torch.tensor(
            lambda_init,
            dtype=self.data_weight.dtype,
        )
        self.data_weight.requires_grad = learnable
        self.fft_centered = fft_centered
        self.fft_normalization = fft_normalization
        self.spatial_dims = spatial_dims if spatial_dims is not None else [-2, -1]

    def forward(
        self, x: torch.Tensor, y: torch.Tensor, sensitivity_maps: torch.Tensor, mask: torch.Tensor
    ) -> torch.Tensor:
        """Forward pass of :class:`DataGDLayer`.

        Parameters
        ----------
        x : torch.Tensor
            Prediction. Shape [batch_size, num_coils, height, width, 2].
        y : torch.Tensor
            Subsampled k-space data. Shape [batch_size, num_coils, height, width, 2].
        sensitivity_maps : torch.Tensor
            Coil sensitivity maps. Shape [batch_size, num_coils, height, width, 2].
        mask : torch.Tensor
            Sampling mask. Shape [batch_size, 1, num_channels, height, width, 1].

        Returns
        -------
        torch.Tensor
            Data loss term.
        """
        A_x_y = (
            torch.sum(
                fft2(
                    complex_mul(x.expand_as(sensitivity_maps), sensitivity_maps),
                    centered=self.fft_centered,
                    normalization=self.fft_normalization,
                    spatial_dims=self.spatial_dims,
                )
                * mask,
                -4,
                keepdim=True,
            )
            - y
        )
        gradD_x = torch.sum(
            complex_mul(
                ifft2(
                    A_x_y * mask,
                    centered=self.fft_centered,
                    normalization=self.fft_normalization,
                    spatial_dims=self.spatial_dims,
                ),
                complex_conj(sensitivity_maps),
            ),
            dim=(-5),
        )
        return x - self.data_weight * gradD_x


class DataProxCGLayer(torch.nn.Module):
    """Solving the prox wrt. data term using Conjugate Gradient as presented in [Aggarwal2018]_.

    References
    ----------
    .. [Aggarwal2018] Aggarwal HK, Mani MP, Jacob M. MoDL: Model-based deep learning architecture for inverse
        problems. IEEE transactions on medical imaging. 2018 Aug 13;38(2):394-405.
    """

    def __init__(
        self,
        lambda_init: float,
        tol: float = 1e-6,
        iterations: int = 10,
        learnable: bool = True,
        fft_centered: bool = False,
        fft_normalization: str = "backward",
        spatial_dims: Optional[Tuple[int, int]] = None,
    ):
        """Inits :class:`DataProxCGLayer`.

        Parameters
        ----------
        lambda_init : float
            Initial value of data term weight lambda.
        tol : float
            Tolerance for the Conjugate Gradient solver. Default is ``1e-6``.
        iterations : int
            Number of iterations for the Conjugate Gradient solver. Default is ``10``.
        learnable : bool
            If True, the data term weight lambda is learnable. Default is ``True``.
        fft_centered : bool
            If True, the FFT is centered. Default is ``False``.
        fft_normalization : str
            FFT normalization. Default is ``"backward"``.
        spatial_dims : tuple of int
            Spatial dimensions of the FFT. Default is ``None``.
        """
        super().__init__()

        self._lambda = torch.nn.Parameter(torch.Tensor(1))
        self._lambda.data = torch.tensor(lambda_init)
        self._lambda_init = lambda_init
        self._lambda.requires_grad = learnable

        self.tol = tol
        self.iter = iterations

        self.op = ConjugateGradient

        self.fft_centered = fft_centered
        self.fft_normalization = fft_normalization
        self.spatial_dims = spatial_dims if spatial_dims is not None else [-2, -1]

    def forward(self, x: torch.Tensor, y: torch.Tensor, sensitivity_maps: torch.Tensor, mask: torch.Tensor):
        """Forward pass of :class:`DataProxCGLayer`.

        Parameters
        ----------
        x : torch.Tensor
            Prediction. Shape [batch_size, num_coils, height, width, 2].
        y : torch.Tensor
            Subsampled k-space data. Shape [batch_size, num_coils, height, width, 2].
        sensitivity_maps : torch.Tensor
            Coil sensitivity maps. Shape [batch_size, num_coils, height, width, 2].
        mask : torch.Tensor
            Sampling mask. Shape [batch_size, 1, num_channels, height, width, 1].

        Returns
        -------
        torch.Tensor
            Data loss term.
        """
        return self.op.apply(
            x,
            self._lambda,
            y,
            sensitivity_maps,
            mask,
            self.tol,
            self.iter,
            self.fft_centered,
            self.fft_normalization,
            self.spatial_dims,
        )

    def set_learnable(self, flag: bool):
        """Set the learnability of the data term weight lambda."""
        self._lambda.requires_grad = flag


class DataVSLayer(torch.nn.Module):
    """DataLayer using variable splitting formulation."""

    def __init__(
        self,
        alpha_init: float = 1.0,
        beta_init: float = 1.0,
        learnable: bool = True,
        fft_centered: bool = False,
        fft_normalization: str = "backward",
        spatial_dims: Optional[Tuple[int, int]] = None,
    ):
        """Inits :class:`DataVSLayer`.

        Parameters
        ----------
        alpha_init : float
            Initial value for the regularization parameter alpha.
        beta_init : float
            Initial value for the regularization parameter beta.
        learnable : bool
            If True, the data term weight lambda is learnable. Default is ``True``.
        fft_centered : bool
            If True, the FFT is centered. Default is ``False``.
        fft_normalization : str
            If "ortho", the FFT is normalized. Default is ``"backward"``.
        spatial_dims : tuple of int
            If not None, the spatial dimensions of the FFT. Default is ``None``.
        """
        super().__init__()
        self.alpha = torch.nn.Parameter(torch.Tensor(1))
        self.alpha.data = torch.tensor(alpha_init, dtype=self.alpha.dtype)

        self.beta = torch.nn.Parameter(torch.Tensor(1))
        self.beta.data = torch.tensor(beta_init, dtype=self.beta.dtype)

        self.learnable = learnable
        self.set_learnable(learnable)

        self.fft_centered = fft_centered
        self.fft_normalization = fft_normalization
        self.spatial_dims = spatial_dims if spatial_dims is not None else [-2, -1]

    def forward(
        self, x: torch.Tensor, y: torch.Tensor, sensitivity_maps: torch.Tensor, mask: torch.Tensor
    ) -> torch.Tensor:
        """Forward pass of :class:`DataVSLayer`.

        Parameters
        ----------
        x : torch.Tensor
            Prediction. Shape [batch_size, num_coils, height, width, 2].
        y : torch.Tensor
            Subsampled k-space data. Shape [batch_size, num_coils, height, width, 2].
        sensitivity_maps : torch.Tensor
            Coil sensitivity maps. Shape [batch_size, num_coils, height, width, 2].
        mask : torch.Tensor
            Sampling mask. Shape [batch_size, 1, num_channels, height, width, 1].

        Returns
        -------
        torch.Tensor
            Data loss term.
        """
        A_x = torch.sum(
            fft2(
                complex_mul(x.expand_as(sensitivity_maps), sensitivity_maps),
                centered=self.fft_centered,
                normalization=self.fft_normalization,
                spatial_dims=self.spatial_dims,
            ),
            -4,
            keepdim=True,
        )
        k_dc = (1 - mask) * A_x + mask * (self.alpha * A_x + (1 - self.alpha) * y)
        x_dc = torch.sum(
            complex_mul(
                ifft2(
                    k_dc,
                    centered=self.fft_centered,
                    normalization=self.fft_normalization,
                    spatial_dims=self.spatial_dims,
                ),
                complex_conj(sensitivity_maps),
            ),
            dim=(-5),
        )
        return self.beta * x + (1 - self.beta) * x_dc

    def set_learnable(self, flag: bool):
        """Set the learnable flag of the parameters.

        Parameters
        ----------
        flag : bool
            If True, the parameters are learnable.
        """
        self.learnable = flag
        self.alpha.requires_grad = self.learnable
        self.beta.requires_grad = self.learnable


class DCLayer(torch.nn.Module):
    """Data Consistency layer from DC-CNN, apply for single coil mainly."""

    def __init__(
        self,
        lambda_init: float,
        learnable: bool = True,
        fft_centered: bool = False,
        fft_normalization: str = "backward",
        spatial_dims: Optional[Tuple[int, int]] = None,
    ):
        """Inits :class:`DCLayer`.

        Parameters
        ----------
        lambda_init : float
            Initial value of data term weight lambda.
        learnable : bool
            If True, the data term weight lambda is learnable. Default is ``True``.
        fft_centered : bool
            If True, the FFT is centered. Default is ``False``.
        fft_normalization : str
            If "ortho", the FFT is normalized. Default is ``"backward"``.
        spatial_dims : tuple of int
            If not None, the spatial dimensions of the FFT. Default is ``None``.
        """
        super().__init__()
        self.lambda_ = torch.nn.Parameter(torch.Tensor(1))
        self.lambda_.data = torch.tensor(lambda_init, dtype=self.lambda_.dtype)

        self.learnable = learnable
        self.set_learnable(learnable)

        self.fft_centered = fft_centered
        self.fft_normalization = fft_normalization
        self.spatial_dims = spatial_dims if spatial_dims is not None else [-2, -1]

    def forward(self, x: torch.Tensor, y: torch.Tensor, mask: torch.Tensor) -> torch.Tensor:
        """Forward pass of :class:`DCLayer`.

        Parameters
        ----------
        x : torch.Tensor
            Prediction. Shape [batch_size, num_coils, height, width, 2].
        y : torch.Tensor
            Subsampled k-space data. Shape [batch_size, num_coils, height, width, 2].
        mask : torch.Tensor
            Sampling mask. Shape [batch_size, 1, num_channels, height, width, 1].

        Returns
        -------
        torch.Tensor
            Data loss term.
        """
        A_x = fft2(x, centered=self.fft_centered, normalization=self.fft_normalization, spatial_dims=self.spatial_dims)
        k_dc = (1 - mask) * A_x + mask * (self.lambda_ * A_x + (1 - self.lambda_) * y)
        return ifft2(
            k_dc, centered=self.fft_centered, normalization=self.fft_normalization, spatial_dims=self.spatial_dims
        )

    def set_learnable(self, flag: bool):
        """Set the learnable flag of the parameters.

        Parameters
        ----------
        flag : bool
            If True, the parameters are learnable.
        """
        self.learnable = flag
        self.lambda_.requires_grad = self.learnable



# coding=utf-8
__author__ = "Dimitris Karkalousos"

# Taken and adapted from: https://github.com/NVIDIA/NeMo/tree/main/nemo/utils

from dataclasses import dataclass
from threading import Lock
from typing import Dict, Optional

from atommic.utils.metaclasses import Singleton


@dataclass()
class ModelMetadataRegistry:
    """A registry for model metadata."""

    guid: str
    gidx: int
    restoration_path: Optional[str] = None


class AppState(metaclass=Singleton):
    """A singleton class that holds the state of the application."""

    def __init__(self):
        """Inits :class:`AppState`."""
        # method call lock
        self.model_parallel_rank = None
        self.__lock = Lock()

        self._app_cfg = None

        # World info
        self._device_id = None
        self._local_rank = None
        self._global_rank = None
        self._model_parallel_rank = None
        self._tensor_model_parallel_rank = None
        self._pipeline_model_parallel_rank = None
        self._data_parallel_rank = None

        self._world_size = None
        self._model_parallel_size = None
        self._tensor_model_parallel_size = None
        self._tensor_model_parallel_group = None
        self._pipeline_model_parallel_size = None
        self._virtual_pipeline_model_parallel_size = None
        self._pipeline_model_parallel_group = None
        self._pipeline_model_parallel_split_rank = None
        self._model_parallel_group = None
        self._data_parallel_size = None
        self._data_parallel_group = None
        self._use_fp8 = False
        self._init_mpi_proc_group = False

        self._random_seed = None

        # Logging info
        self._log_dir = None
        self._exp_dir = None
        self._name = None
        self._checkpoint_name = None
        self._version = None
        self._create_checkpoint_callback = None
        self._checkpoint_callback_params = None

        # Save and Restore (.atommic)
        self._tmpdir_name = None
        self._is_model_being_restored = False
        self._atommic_file_folder = None
        self._model_restore_path = None
        self._all_model_restore_paths = []
        self._model_guid_map = {}  # type: Dict[str, ModelMetadataRegistry]

    @property
    def device_id(self):
        """Property returns the device_id."""
        return self._device_id

    @device_id.setter
    def device_id(self, id: int):  # pylint: disable=redefined-builtin
        """Property sets the device_id."""
        self._device_id = id

    @property
    def world_size(self):
        """Property returns the total number of GPUs."""
        return self._world_size

    @world_size.setter
    def world_size(self, size: int):
        """Property sets the total number of GPUs."""
        self._world_size = size

    @property
    def model_parallel_size(self):
        """Property returns the number of GPUs in each model parallel group."""
        return self._model_parallel_size

    @model_parallel_size.setter
    def model_parallel_size(self, size: int):
        """Property sets the number of GPUs in each model parallel group."""
        self._model_parallel_size = size

    @property
    def tensor_model_parallel_size(self):
        """Property returns the number of GPUs in each model parallel group."""
        return self._tensor_model_parallel_size

    @tensor_model_parallel_size.setter
    def tensor_model_parallel_size(self, size):
        """Property sets the number of GPUs in each model parallel group."""
        self._tensor_model_parallel_size = size

    @property
    def pipeline_model_parallel_size(self):
        """Property returns the number of GPUs in each model parallel group."""
        return self._pipeline_model_parallel_size

    @pipeline_model_parallel_size.setter
    def pipeline_model_parallel_size(self, size):
        """Property sets the number of GPUs in each model parallel group."""
        self._pipeline_model_parallel_size = size

    @property
    def virtual_pipeline_model_parallel_size(self):
        """Property returns the number of GPUs in each model parallel group."""
        return self._virtual_pipeline_model_parallel_size

    @virtual_pipeline_model_parallel_size.setter
    def virtual_pipeline_model_parallel_size(self, size):
        """Property sets the size of the virtual pipeline parallel model."""
        self._virtual_pipeline_model_parallel_size = size

    @property
    def data_parallel_size(self):
        """Property returns the number of GPUs in each data parallel group."""
        return self._data_parallel_size

    @data_parallel_size.setter
    def data_parallel_size(self, size: int):
        """Property sets the number of GPUs in each data parallel group."""
        self._data_parallel_size = size

    @property
    def local_rank(self):
        """Property returns the local rank."""
        return self._local_rank

    @local_rank.setter
    def local_rank(self, rank: int):
        """Property sets the local rank."""
        self._local_rank = rank

    @property
    def global_rank(self):
        """Property returns the global rank."""
        return self._global_rank

    @global_rank.setter
    def global_rank(self, rank: int):
        """Property sets the global rank."""
        self._global_rank = rank

    @property
    def tensor_model_parallel_rank(self):
        """Property returns the model parallel rank."""
        return self._tensor_model_parallel_rank

    @tensor_model_parallel_rank.setter
    def tensor_model_parallel_rank(self, rank):
        """Property sets the model parallel rank."""
        self._tensor_model_parallel_rank = rank

    @property
    def tensor_model_parallel_group(self):
        """Property returns the model parallel group."""
        return self._tensor_model_parallel_group

    @tensor_model_parallel_group.setter
    def tensor_model_parallel_group(self, group):
        """Property sets the model parallel group."""
        self._tensor_model_parallel_group = group

    @property
    def pipeline_model_parallel_rank(self):
        """Property returns the model parallel rank."""
        return self._pipeline_model_parallel_rank

    @pipeline_model_parallel_rank.setter
    def pipeline_model_parallel_rank(self, rank):
        """Property sets the model parallel rank."""
        self._pipeline_model_parallel_rank = rank

    @property
    def virtual_pipeline_model_parallel_rank(self):
        """Property returns the virtual pipeline parallel rank."""
        return self._virtual_pipeline_model_parallel_rank

    @virtual_pipeline_model_parallel_rank.setter
    def virtual_pipeline_model_parallel_rank(self, rank):
        """Property sets the virtual pipeline parallel rank."""
        self._virtual_pipeline_model_parallel_rank = rank

    @property
    def pipeline_model_parallel_split_rank(self):
        """Property returns the model parallel split rank."""
        return self._pipeline_model_parallel_split_rank

    @pipeline_model_parallel_split_rank.setter
    def pipeline_model_parallel_split_rank(self, rank):
        """Property sets the model parallel split rank."""
        self._pipeline_model_parallel_split_rank = rank

    @property
    def pipeline_model_parallel_group(self):
        """Property returns the model parallel group."""
        return self._pipeline_model_parallel_group

    @pipeline_model_parallel_group.setter
    def pipeline_model_parallel_group(self, group):
        """Property sets the model parallel group."""
        self._pipeline_model_parallel_group = group

    @property
    def data_parallel_rank(self):
        """Property returns the data parallel rank."""
        return self._data_parallel_rank

    @data_parallel_rank.setter
    def data_parallel_rank(self, rank: int):
        """Property sets the data parallel rank."""
        self._data_parallel_rank = rank

    @property
    def data_parallel_group(self):
        """Property returns the data parallel group."""
        return self._data_parallel_group

    @data_parallel_group.setter
    def data_parallel_group(self, group):
        """Property sets the data parallel group."""
        self._data_parallel_group = group

    @property
    def use_fp8(self):
        """Property returns the use of fp8 precision.

        Returns
        -------
            Use of FP8.
        """
        return self._use_fp8

    @use_fp8.setter
    def use_fp8(self, use_fp8):
        """Property sets the use of fp8 precision.

        Parameters
        ----------
        use_fp8 : bool
            Use of FP8.
        """
        self._use_fp8 = use_fp8

    @property
    def init_mpi_proc_group(self):
        """Property sets the initialization of mpi process group.

        Returns
        -------
            Initialize mpi process group.
        """
        return self._init_mpi_proc_group

    @init_mpi_proc_group.setter
    def init_mpi_proc_group(self, init_mpi_proc_group):
        """Property sets the initialization of mpi process group.

        Parameters
        ----------
        init_mpi_proc_group : bool
            Initialize mpi process group.
        """
        self._init_mpi_proc_group = init_mpi_proc_group

    @property
    def random_seed(self):
        """Property returns the random seed."""
        return self._random_seed

    @random_seed.setter
    def random_seed(self, seed: int):
        """Property sets the random seed."""
        self._random_seed = seed

    @property
    def log_dir(self):
        """Returns the log_dir set by exp_manager."""
        return self._log_dir

    @log_dir.setter
    def log_dir(self, dir):  # pylint: disable=redefined-builtin
        """Sets the log_dir property."""
        self._log_dir = dir

    @property
    def exp_dir(self):
        """Returns the exp_dir set by exp_manager."""
        return self._exp_dir

    @exp_dir.setter
    def exp_dir(self, dir):  # pylint: disable=redefined-builtin
        """Sets the log_dir property."""
        self._exp_dir = dir

    @property
    def name(self):
        """Returns the name set by exp_manager."""
        return self._name

    @name.setter
    def name(self, name: str):
        """Sets the name property."""
        self._name = name

    @property
    def checkpoint_name(self):
        """Returns the name set by exp_manager."""
        return self._checkpoint_name

    @checkpoint_name.setter
    def checkpoint_name(self, name: str):
        """Sets the name property."""
        self._checkpoint_name = name

    @property
    def version(self):
        """Returns the version set by exp_manager."""
        return self._version

    @version.setter
    def version(self, version: str):
        """Sets the version property."""
        self._version = version

    @property
    def create_checkpoint_callback(self):
        """Returns the create_checkpoint_callback set by exp_manager."""
        return self._create_checkpoint_callback

    @create_checkpoint_callback.setter
    def create_checkpoint_callback(self, create_checkpoint_callback: bool):
        """Sets the create_checkpoint_callback property."""
        self._create_checkpoint_callback = create_checkpoint_callback

    @property
    def checkpoint_callback_params(self):
        """Returns the version set by exp_manager."""
        return self._checkpoint_callback_params

    @checkpoint_callback_params.setter
    def checkpoint_callback_params(self, params: dict):
        """Sets the name property."""
        self._checkpoint_callback_params = params

    @property
    def model_restore_path(self):
        """Returns the model_restore_path set by exp_manager."""
        return self._all_model_restore_paths[-1] if len(self._all_model_restore_paths) > 0 else None

    @model_restore_path.setter
    def model_restore_path(self, path):
        """Sets the model_restore_path property."""
        with self.__lock:
            self._model_restore_path = path
            self._all_model_restore_paths.append(path)

    def register_model_guid(self, guid: str, restoration_path: Optional[str] = None):
        """Maps a guid to its restore path (None or last absolute path)."""
        with self.__lock:
            if guid in self._model_guid_map:
                idx = self._model_guid_map[guid].gidx
            else:
                idx = len(self._model_guid_map)
            self._model_guid_map[guid] = ModelMetadataRegistry(guid, idx, restoration_path=restoration_path)

    def reset_model_guid_registry(self):
        """Resets the model guid registry."""
        with self.__lock:
            self._model_guid_map.clear()

    def get_model_metadata_from_guid(self, guid) -> ModelMetadataRegistry:
        """Returns the global model idx and restoration path."""
        return self._model_guid_map[guid]

    @property
    def is_model_being_restored(self) -> bool:
        """Returns whether a model is being restored."""
        return self._is_model_being_restored

    @is_model_being_restored.setter
    def is_model_being_restored(self, is_restored: bool):
        """Sets whether a model is being restored."""
        self._is_model_being_restored = is_restored

    @property
    def atommic_file_folder(self) -> str:
        """Returns the atommic_file_folder set by exp_manager."""
        return self._atommic_file_folder

    @atommic_file_folder.setter
    def atommic_file_folder(self, path: str):
        """Sets the atommic_file_folder property."""
        self._atommic_file_folder = path



# coding=utf-8
__author__ = "Dimitris Karkalousos"

import math
from typing import Dict, List, Optional, Tuple, Union

import torch

from atommic.collections.common.parts.utils import rnn_weights_init
from atommic.collections.reconstruction.nn.rim_base import rim_block
from atommic.collections.reconstruction.nn.rim_base.conv_layers import ConvNonlinear
from atommic.collections.reconstruction.nn.unet_base.unet_block import Unet
from atommic.collections.segmentation.nn.attentionunet_base.attentionunet_block import AttentionUnet
from atommic.collections.segmentation.nn.lambdaunet_base.lambdaunet_block import LambdaBlock
from atommic.collections.segmentation.nn.vnet_base.vnet_block import VNet

__all__ = ["MTLRSBlock"]


class MTLRSBlock(torch.nn.Module):
    """Implementation of a Multi-Task Learning for MRI Reconstruction and Segmentation (MTLRS) block, as presented in
    [Karkalousos2023]_.

    References
    ----------
    .. [Karkalousos2023] Karkalousos, D., Išgum, I., Marquering, H., Caan, M. W. A., (2023). MultiTask Learning for
        accelerated-MRI Reconstruction and Segmentation of Brain Lesions in Multiple Sclerosis. In Proceedings of
        Machine Learning Research (Vol. 078).
    """

    def __init__(
        self,
        reconstruction_module_params: Dict,
        segmentation_module_params: Dict,
        input_channels: int,
        magnitude_input: bool = True,
        fft_centered: bool = False,
        fft_normalization: str = "backward",
        spatial_dims: Optional[Tuple[int, int]] = None,
        coil_dim: int = 1,
        dimensionality: int = 2,
        consecutive_slices: int = 1,
        coil_combination_method: str = "SENSE",
        normalize_segmentation_output: bool = True,
    ):
        """Inits :class:`MTLRSBlock`.

        Parameters
        ----------
        reconstruction_module_params : Dict
            Parameters for the reconstruction module.
        segmentation_module_params : Dict
            Parameters for the segmentation module.
        input_channels : int
            Number of input channels.
        magnitude_input : bool
            Whether the input is magnitude or complex. Default is ``True``.
        fft_centered : bool
            Whether the FFT is centered. Default is ``False``.
        fft_normalization : str
            Normalization of the FFT. Default is ``"backward"``.
        spatial_dims : Tuple[int, int]
            Spatial dimensions of the input. Default is ``None``.
        coil_dim : int
            Coil dimension of the input. Default is ``1``.
        dimensionality : int
            Dimensionality of the input. Default is ``2``.
        consecutive_slices : int
            Number of consecutive slices to be used. Default is ``1``.
        coil_combination_method : str
            Coil combination method. Default is ``"SENSE"``.
        normalize_segmentation_output : bool
            Whether to normalize the segmentation output. Default is ``True`` .
        """
        super().__init__()

        # General parameters
        self.input_channels = input_channels
        self.magnitude_input = magnitude_input
        self.fft_centered = fft_centered
        self.fft_normalization = fft_normalization
        self.spatial_dims = spatial_dims
        self.coil_dim = coil_dim
        self.dimensionality = dimensionality
        if self.dimensionality != 2:
            raise NotImplementedError(f"Currently only 2D is supported for segmentation, got {self.dimensionality}D.")
        self.consecutive_slices = consecutive_slices
        self.coil_combination_method = coil_combination_method

        # Reconstruction module parameters
        self.reconstruction_module_params = reconstruction_module_params
        self.reconstruction_module_recurrent_filters = self.reconstruction_module_params["recurrent_filters"]
        self.reconstruction_module_time_steps = 8 * math.ceil(self.reconstruction_module_params["time_steps"] / 8)
        self.no_dc = self.reconstruction_module_params["no_dc"]
        self.keep_prediction = self.reconstruction_module_params["keep_prediction"]
        self.reconstruction_module_dimensionality = self.reconstruction_module_params["dimensionality"]
        reconstruction_module_consecutive_slices = (
            self.consecutive_slices if self.reconstruction_module_dimensionality == 3 else 1
        )
        self.reconstruction_module = torch.nn.ModuleList(
            [
                rim_block.RIMBlock(
                    recurrent_layer=self.reconstruction_module_params["recurrent_layer"],
                    conv_filters=self.reconstruction_module_params["conv_filters"],
                    conv_kernels=self.reconstruction_module_params["conv_kernels"],
                    conv_dilations=self.reconstruction_module_params["conv_dilations"],
                    conv_bias=self.reconstruction_module_params["conv_bias"],
                    recurrent_filters=self.reconstruction_module_recurrent_filters,
                    recurrent_kernels=self.reconstruction_module_params["recurrent_kernels"],
                    recurrent_dilations=self.reconstruction_module_params["recurrent_dilations"],
                    recurrent_bias=self.reconstruction_module_params["recurrent_bias"],
                    depth=self.reconstruction_module_params["depth"],
                    time_steps=self.reconstruction_module_time_steps,
                    conv_dim=self.reconstruction_module_params["conv_dim"],
                    no_dc=self.no_dc,
                    fft_centered=self.fft_centered,
                    fft_normalization=self.fft_normalization,
                    spatial_dims=self.spatial_dims,
                    coil_dim=self.coil_dim - 1,
                    dimensionality=self.reconstruction_module_dimensionality,
                    consecutive_slices=reconstruction_module_consecutive_slices,
                    coil_combination_method=self.coil_combination_method,
                )
                for _ in range(self.reconstruction_module_params["num_cascades"])
            ]
        )
        # Keep estimation through the cascades if keep_prediction is True or re-estimate it if False.
        self.reconstruction_module_keep_prediction = self.reconstruction_module_params["keep_prediction"]
        # initialize weights if not using pretrained cirim
        if not self.reconstruction_module_params["pretrained"]:
            std_init_range = 1 / self.reconstruction_module_recurrent_filters[0] ** 0.5
            self.reconstruction_module.apply(lambda module: rnn_weights_init(module, std_init_range))
        self.dc_weight = torch.nn.Parameter(torch.ones(1))
        self.accumulate_predictions = self.reconstruction_module_params["accumulate_predictions"]

        # Segmentation module parameters
        self.segmentation_module_params = segmentation_module_params
        segmentation_module = self.segmentation_module_params["segmentation_module"]
        self.segmentation_module_output_channels = self.segmentation_module_params["output_channels"]
        if segmentation_module.lower() == "unet":
            segmentation_module = Unet(
                in_chans=self.input_channels,
                out_chans=self.segmentation_module_output_channels,
                chans=self.segmentation_module_params["channels"],
                num_pool_layers=self.segmentation_module_params["pooling_layers"],
                drop_prob=self.segmentation_module_params["dropout"],
            )
        elif segmentation_module.lower() == "attentionunet":
            segmentation_module = AttentionUnet(
                in_chans=self.input_channels,
                out_chans=self.segmentation_module_output_channels,
                chans=self.segmentation_module_params["channels"],
                num_pool_layers=self.segmentation_module_params["pooling_layers"],
                drop_prob=self.segmentation_module_params["dropout"],
            )
        elif segmentation_module.lower() == "lambdaunet":
            segmentation_module = LambdaBlock(
                in_chans=self.input_channels,
                out_chans=self.segmentation_module_output_channels,
                drop_prob=self.segmentation_module_params["dropout"],
                temporal_kernel=self.segmentation_module_params["temporal_kernel"],
                num_slices=self.consecutive_slices,
            )
        elif segmentation_module.lower() == "vnet":
            segmentation_module = VNet(
                in_chans=self.input_channels,
                out_chans=self.segmentation_module_output_channels,
                act=self.segmentation_module_params["activation"],
                drop_prob=self.segmentation_module_params["dropout"],
                bias=self.segmentation_module_params["bias"],
            )
        elif segmentation_module.lower() == "convlayer":
            segmentation_module = torch.nn.Sequential(
                ConvNonlinear(
                    self.input_channels,
                    self.segmentation_module_output_channels,
                    conv_dim=self.segmentation_module_params["conv_dim"],
                    kernel_size=3,
                    dilation=1,
                    bias=False,
                    nonlinear=None,  # No nonlinear activation
                )
            )
        else:
            raise ValueError(f"Segmentation module {segmentation_module} not implemented.")
        self.segmentation_module = segmentation_module

        self.normalize_segmentation_output = normalize_segmentation_output

    def forward(  # noqa: MC0001
        self,
        y: torch.Tensor,
        sensitivity_maps: torch.Tensor,
        mask: torch.Tensor,
        init_reconstruction_pred: torch.Tensor,
        target_reconstruction: torch.Tensor,  # pylint: disable=unused-argument
        hx: torch.Tensor = None,
        sigma: float = 1.0,
    ) -> Tuple[Union[List, torch.Tensor], torch.Tensor]:
        """Forward pass of :class:`MTLRSBlock`.

        Parameters
        ----------
        y : torch.Tensor
            Subsampled k-space data. Shape [batch_size, n_coils, n_x, n_y, 2]
        sensitivity_maps : torch.Tensor
            Coil sensitivity maps. Shape [batch_size, n_coils, n_x, n_y, 2]
        mask : torch.Tensor
            Subsampling mask. Shape [1, 1, n_x, n_y, 1]
        init_reconstruction_pred : torch.Tensor
            Initial reconstruction prediction. Shape [batch_size, n_x, n_y, 2]
        target_reconstruction : torch.Tensor
            Target reconstruction. Shape [batch_size, n_x, n_y, 2]
        hx : torch.Tensor, optional
            Initial hidden state for the RNN. Default is ``None``.
        sigma : float, optional
            Standard deviation of the noise. Default is ``1.0``.

        Returns
        -------
        Tuple[Union[List, torch.Tensor], torch.Tensor]
            Tuple containing the predicted reconstruction and segmentation.
        """
        if self.consecutive_slices > 1 and self.reconstruction_module_dimensionality == 2:
            # Do per slice reconstruction
            pred_reconstruction_slices = []
            for slice_idx in range(self.consecutive_slices):
                y_slice = y[:, slice_idx, ...]
                prediction_slice = y_slice.clone()
                sensitivity_maps_slice = sensitivity_maps[:, slice_idx, ...]
                mask_slice = mask[:, 0, ...]
                init_reconstruction_pred_slice = init_reconstruction_pred[:, slice_idx, ...]
                _pred_reconstruction_slice = (
                    None
                    if init_reconstruction_pred_slice is None or init_reconstruction_pred_slice.dim() < 4
                    else init_reconstruction_pred_slice
                )
                cascades_predictions = []
                for i, cascade in enumerate(self.reconstruction_module):
                    # Forward pass through the cascades
                    prediction_slice, hx = cascade(
                        prediction_slice,
                        y_slice,
                        sensitivity_maps_slice,
                        mask_slice,
                        _pred_reconstruction_slice,
                        hx,
                        sigma,
                        keep_prediction=False if i == 0 else self.keep_prediction,
                    )
                    time_steps_predictions = [torch.view_as_complex(pred) for pred in prediction_slice]
                    cascades_predictions.append(torch.stack(time_steps_predictions, dim=0))
                pred_reconstruction_slices.append(torch.stack(cascades_predictions, dim=0))
            preds = torch.stack(pred_reconstruction_slices, dim=3)

            cascades_predictions = [
                [
                    preds[cascade_prediction, time_step_prediction, ...]
                    for time_step_prediction in range(preds.shape[1])
                ]
                for cascade_prediction in range(preds.shape[0])
            ]
        else:
            prediction = y.clone()
            _pred_reconstruction = (
                None
                if init_reconstruction_pred is None or init_reconstruction_pred.dim() < 4
                else init_reconstruction_pred
            )
            sigma = 1.0
            cascades_predictions = []
            for i, cascade in enumerate(self.reconstruction_module):
                # Forward pass through the cascades
                prediction, hx = cascade(
                    prediction,
                    y,
                    sensitivity_maps,
                    mask,
                    _pred_reconstruction,
                    hx,
                    sigma,
                    keep_prediction=False if i == 0 else self.keep_prediction,
                )
                time_steps_predictions = [torch.view_as_complex(pred) for pred in prediction]
                cascades_predictions.append(time_steps_predictions)
        pred_reconstruction = cascades_predictions

        _pred_reconstruction = pred_reconstruction
        if isinstance(_pred_reconstruction, list):
            _pred_reconstruction = _pred_reconstruction[-1]
        if isinstance(_pred_reconstruction, list):
            _pred_reconstruction = _pred_reconstruction[-1]
        if _pred_reconstruction.shape[-1] != 2:
            _pred_reconstruction = torch.view_as_real(_pred_reconstruction)
        if self.consecutive_slices > 1 and _pred_reconstruction.dim() == 5:
            _pred_reconstruction = _pred_reconstruction.reshape(
                _pred_reconstruction.shape[0] * _pred_reconstruction.shape[1],
                *_pred_reconstruction.shape[2:],
            )
        if _pred_reconstruction.shape[-1] == 2:
            if self.input_channels == 1:
                _pred_reconstruction = torch.view_as_complex(_pred_reconstruction).unsqueeze(1)
                if self.magnitude_input:
                    _pred_reconstruction = torch.abs(_pred_reconstruction)
            elif self.input_channels == 2:
                if self.magnitude_input:
                    raise ValueError("Magnitude input is not supported for 2-channel input.")
                _pred_reconstruction = _pred_reconstruction.permute(0, 3, 1, 2)
            else:
                raise ValueError(f"The input channels must be either 1 or 2. Found: {self.input_channels}")
        else:
            _pred_reconstruction = _pred_reconstruction.unsqueeze(1)

        pred_segmentation = self.segmentation_module(torch.abs(_pred_reconstruction))

        if self.normalize_segmentation_output:
            pred_segmentation = (pred_segmentation - pred_segmentation.min()) / (
                pred_segmentation.max() - pred_segmentation.min()
            )

        pred_segmentation = torch.abs(pred_segmentation)

        if self.consecutive_slices > 1:
            # get batch size and number of slices from y, because if the reconstruction module is used they will
            # not be saved before
            pred_segmentation = pred_segmentation.view([y.shape[0], y.shape[1], *pred_segmentation.shape[1:]])

        return pred_reconstruction, pred_segmentation, hx  # type: ignore



# coding=utf-8
__author__ = "Dimitris Karkalousos"

# Taken and adapted from: https://github.com/NVIDIA/NeMo/blob/main/nemo/core/neural_types/axes.py

from enum import Enum
from typing import Optional

__all__ = ["AxisKindAbstract", "AxisKind", "AxisType"]


class AxisKindAbstract(Enum):
    """This is an abstract Enum to represents what does varying axis dimension mean. In practice, you will almost
    always use AxisKind Enum. This Enum should be inherited by your OWN Enum if you aren't satisfied with AxisKind.
    Then your own Enum can be used instead of AxisKind.
    """


class AxisKind(AxisKindAbstract):
    """This Enum represents what does varying axis dimension mean. For example, does this dimension correspond to
    width, batch, time, etc. The "Dimension" and "Channel" kinds are the same and used to represent a general axis.
    "Any" axis will accept any axis kind fed to it.
    """

    Batch = 0
    Time = 1
    Dimension = 2
    Width = 3
    Height = 4
    Any = 5
    Sequence = 6
    FlowGroup = 7
    Singleton = 8  # Used to represent a axis that has size 1

    def __repr__(self):
        """Returns short string representation of the AxisKind"""
        return self.__str__()

    def __str__(self):
        """Returns short string representation of the AxisKind"""
        return str(self.name).lower()

    @staticmethod
    def from_str(label):
        """Returns AxisKind instance based on short string representation"""
        _label = label.lower().strip()
        if _label in ("b", "n", "batch"):
            return AxisKind.Batch
        if _label == "t" or _label == "time" or (len(_label) > 2 and _label.startswith("t_")):
            return AxisKind.Time
        if _label in ("d", "c", "channel"):
            return AxisKind.Dimension
        if _label in ("w", "width"):
            return AxisKind.Width
        if _label in ("h", "height"):
            return AxisKind.Height
        if _label in ("s", "singleton"):
            return AxisKind.Singleton
        if _label in ("seq", "sequence"):
            return AxisKind.Sequence
        if _label == "flowgroup":
            return AxisKind.FlowGroup
        if _label == "any":
            return AxisKind.Any
        raise ValueError(f"Can't create AxisKind from {label}")


class AxisType:
    """This class represents axis semantics and (optionally) it's dimensionality."""

    def __init__(self, kind: AxisKindAbstract, size: Optional[int] = None, is_list=False):
        """Inits :class:`AxisType`.

        Parameters
        ----------
        kind : AxisKindAbstract
            What kind of axis it is? For example Batch, Height, etc.
        size : int, optional
            Specify if the axis should have a fixed size. By default, it is set to None and you typically do not want
            to set it for Batch and Time. Default is ``None``.
        is_list : bool, optional
            Specify if the axis is a list. Default is ``False``.
        """
        if size is not None and is_list:
            raise ValueError("The axis can't be list and have a fixed size")
        self.kind = kind
        self.size = size
        self.is_list = is_list

    def __repr__(self):
        """Returns short string representation of the AxisType"""
        if self.size is None:
            representation = str(self.kind)
        else:
            representation = f"{str(self.kind)}:{self.size}"
        if self.is_list:
            representation += "_listdim"
        return representation



# coding=utf-8
__author__ = "Dimitris Karkalousos"

# Parts of the code have been taken from https://github.com/facebookresearch/fastMRI

import math
from typing import List, Tuple

import torch


class NormUnet(torch.nn.Module):
    """Normalized U-Net model.

    This is the same as a regular U-Net, but with normalization applied to the input before the U-Net. This keeps the
    values more numerically stable during training.
    """

    def __init__(
        self,
        chans: int,
        num_pools: int,
        in_chans: int = 2,
        out_chans: int = 2,
        drop_prob: float = 0.0,
        padding_size: int = 15,
        normalize: bool = True,
        norm_groups: int = 2,
    ):
        """Inits :class:`NormUnet`.

        Parameters
        ----------
        chans : int
            Number of output channels of the first convolution layer.
        num_pools : int
            Number of down-sampling and up-sampling layers.
        in_chans : int, optional
            Number of channels in the input to the U-Net model. Default is ``2``.
        out_chans : int, optional
            Number of channels in the output to the U-Net model. Default is ``2``.
        drop_prob : float, optional
            Dropout probability. Default is ``0.0``.
        padding_size : int, optional
            Size of the padding. Default is ``15``.
        normalize : bool, optional
            Whether to normalize the input. Default is ``True``.
        norm_groups : int, optional
            Number of groups to use for group normalization. Default is ``2``.
        """
        super().__init__()
        self.unet = Unet(
            in_chans=in_chans, out_chans=out_chans, chans=chans, num_pool_layers=num_pools, drop_prob=drop_prob
        )
        self.padding_size = padding_size
        self.normalize = normalize
        self.norm_groups = norm_groups

    @staticmethod
    def complex_to_chan_dim(x: torch.Tensor) -> torch.Tensor:
        """Convert the last dimension of the input to complex."""
        b, c, h, w, two = x.shape
        if two != 2:
            raise AssertionError
        return x.permute(0, 4, 1, 2, 3).reshape(b, 2 * c, h, w)

    @staticmethod
    def chan_complex_to_last_dim(x: torch.Tensor) -> torch.Tensor:
        """Convert the last dimension of the input to complex."""
        b, c2, h, w = x.shape
        if c2 % 2 != 0:
            raise AssertionError
        c = torch.div(c2, 2, rounding_mode="trunc")
        return x.view(b, 2, c, h, w).permute(0, 2, 3, 4, 1).contiguous()

    def norm(self, x: torch.Tensor) -> Tuple[torch.Tensor, torch.Tensor, torch.Tensor]:
        """Normalize the input."""
        # group norm
        b, c, h, w = x.shape

        x = x.reshape(b, self.norm_groups, -1)

        mean = x.mean(-1, keepdim=True)
        std = x.std(-1, keepdim=True)

        x = (x - mean) / std

        x = x.reshape(b, c, h, w)

        return x, mean, std

    def unnorm(self, x: torch.Tensor, mean: torch.Tensor, std: torch.Tensor) -> torch.Tensor:
        """Unnormalize the input."""
        b, c, h, w = x.shape
        input_data = x.reshape(b, self.norm_groups, -1)
        return (input_data * std + mean).reshape(b, c, h, w)

    def pad(self, x: torch.Tensor) -> Tuple[torch.Tensor, Tuple[List[int], List[int], int, int]]:
        """Pad the input with zeros to make it square."""
        _, _, h, w = x.shape
        w_mult = ((w - 1) | self.padding_size) + 1
        h_mult = ((h - 1) | self.padding_size) + 1
        w_pad = [math.floor((w_mult - w) / 2), math.ceil((w_mult - w) / 2)]
        h_pad = [math.floor((h_mult - h) / 2), math.ceil((h_mult - h) / 2)]
        # TODO: fix this type when PyTorch fixes theirs
        # the documentation lies - this actually takes a list
        # https://github.com/pytorch/pytorch/blob/master/torch/nn/functional.py#L3457
        # https://github.com/pytorch/pytorch/pull/16949
        x = torch.nn.functional.pad(x, w_pad + h_pad)

        return x, (h_pad, w_pad, h_mult, w_mult)

    @staticmethod
    def unpad(x: torch.Tensor, h_pad: List[int], w_pad: List[int], h_mult: int, w_mult: int) -> torch.Tensor:
        """Unpad the input."""
        return x[..., h_pad[0] : h_mult - h_pad[1], w_pad[0] : w_mult - w_pad[1]]

    def forward(self, x: torch.Tensor) -> torch.Tensor:
        """Forward pass of :class:`NormUnet`."""
        iscomplex = False
        if x.shape[-1] == 2:
            x = self.complex_to_chan_dim(x)
            iscomplex = True

        mean = 1.0
        std = 1.0

        if self.normalize:
            x, mean, std = self.norm(x)

        if self.padding_size > 0:
            x, pad_sizes = self.pad(x)

        x = self.unet(x)

        if self.padding_size > 0:
            x = self.unpad(x, *pad_sizes)

        if self.normalize:
            x = self.unnorm(x, mean, std)

        if iscomplex:
            x = self.chan_complex_to_last_dim(x)

        return x


class Unet(torch.nn.Module):
    """U-Net model, as presented in [Ronneberger2015]_.

    References
    ----------
    .. [Ronneberger2015] O. Ronneberger, P. Fischer, and Thomas Brox. U-net: Convolutional networks for biomedical
        image segmentation. In International Conference on Medical image computing and computer-assisted intervention,
        pages 234–241. Springer, 2015.
    """

    def __init__(
        self, in_chans: int, out_chans: int, chans: int = 32, num_pool_layers: int = 4, drop_prob: float = 0.0
    ):
        """Inits :class:`Unet`.

        Parameters
        ----------
        in_chans : int
            Number of channels in the input to the U-Net model.
        out_chans : int
            Number of channels in the output to the U-Net model.
        chans : int
            Number of output channels of the first convolution layer. Default is ``32``.
        num_pool_layers : int
            Number of down-sampling and up-sampling layers. Default is ``4``.
        drop_prob : float
            Dropout probability. Default is ``0.0``.
        """
        super().__init__()

        self.in_chans = in_chans
        self.out_chans = out_chans
        self.chans = chans
        self.num_pool_layers = num_pool_layers
        self.drop_prob = drop_prob

        self.down_sample_layers = torch.nn.ModuleList([ConvBlock(in_chans, chans, drop_prob)])
        ch = chans
        for _ in range(num_pool_layers - 1):
            self.down_sample_layers.append(ConvBlock(ch, ch * 2, drop_prob))
            ch = ch * 2
        self.conv = ConvBlock(ch, ch * 2, drop_prob)

        self.up_conv = torch.nn.ModuleList()
        self.up_transpose_conv = torch.nn.ModuleList()
        for _ in range(num_pool_layers - 1):
            self.up_transpose_conv.append(TransposeConvBlock(ch * 2, ch))
            self.up_conv.append(ConvBlock(ch * 2, ch, drop_prob))
            ch = ch // 2

        self.up_transpose_conv.append(TransposeConvBlock(ch * 2, ch))
        self.up_conv.append(
            torch.nn.Sequential(
                ConvBlock(ch * 2, ch, drop_prob), torch.nn.Conv2d(ch, self.out_chans, kernel_size=1, stride=1)
            )
        )

    def forward(self, image: torch.Tensor) -> torch.Tensor:
        """Forward pass of :class:`Unet`.

        Parameters
        ----------
        image : torch.Tensor
            Input tensor of shape `(N, in_chans, H, W)`.

        Returns
        -------
        torch.Tensor
            Output tensor of shape `(N, out_chans, H, W)`.
        """
        stack = []
        output = image

        # apply down-sampling layers
        for layer in self.down_sample_layers:
            output = layer(output)
            stack.append(output)
            output = torch.nn.functional.avg_pool2d(output, kernel_size=2, stride=2, padding=0)

        output = self.conv(output)

        # apply up-sampling layers
        for transpose_conv, conv in zip(self.up_transpose_conv, self.up_conv):
            downsample_layer = stack.pop()
            output = transpose_conv(output)

            # reflect pad on the right/bottom if needed to handle odd input dimensions
            padding = [0, 0, 0, 0]
            if output.shape[-1] != downsample_layer.shape[-1]:
                padding[1] = 1  # padding right
            if output.shape[-2] != downsample_layer.shape[-2]:
                padding[3] = 1  # padding bottom
            if torch.sum(torch.tensor(padding)) != 0:
                output = torch.nn.functional.pad(output, padding, "reflect")

            output = torch.cat([output, downsample_layer], dim=1)
            output = conv(output)

        return output


class ConvBlock(torch.nn.Module):
    """A Convolutional Block that consists of two convolution layers each followed by instance normalization, LeakyReLU
    activation and dropout.
    """

    def __init__(self, in_chans: int, out_chans: int, drop_prob: float):
        """Inits :class:`ConvBlock`.

        Parameters
        ----------
        in_chans : int
            Number of channels in the input.
        out_chans : int
            Number of channels in the output.
        drop_prob : float
            Dropout probability.
        """
        super().__init__()

        self.in_chans = in_chans
        self.out_chans = out_chans
        self.drop_prob = drop_prob

        self.layers = torch.nn.Sequential(
            torch.nn.Conv2d(in_chans, out_chans, kernel_size=3, padding=1, bias=False),
            torch.nn.InstanceNorm2d(out_chans),
            torch.nn.LeakyReLU(negative_slope=0.2, inplace=True),
            torch.nn.Dropout2d(drop_prob),
            torch.nn.Conv2d(out_chans, out_chans, kernel_size=3, padding=1, bias=False),
            torch.nn.InstanceNorm2d(out_chans),
            torch.nn.LeakyReLU(negative_slope=0.2, inplace=True),
            torch.nn.Dropout2d(drop_prob),
        )

    def forward(self, image: torch.Tensor) -> torch.Tensor:
        """Forward pass of :class:`ConvBlock`.

        Parameters
        ----------
        image : torch.Tensor
            Input tensor of shape `(N, in_chans, H, W)`.

        Returns
        -------
        torch.Tensor
            Output tensor of shape `(N, out_chans, H, W)`.
        """
        return self.layers(image)


class TransposeConvBlock(torch.nn.Module):
    """A Transpose Convolutional Block that consists of one convolution transpose layers followed by instance
    normalization and LeakyReLU activation.
    """

    def __init__(self, in_chans: int, out_chans: int):
        """Inits :class:`TransposeConvBlock`.

        Parameters
        ----------
        in_chans : int
            Number of channels in the input.
        out_chans : int
            Number of channels in the output.
        """
        super().__init__()

        self.in_chans = in_chans
        self.out_chans = out_chans

        self.layers = torch.nn.Sequential(
            torch.nn.ConvTranspose2d(in_chans, out_chans, kernel_size=2, stride=2, bias=False),
            torch.nn.InstanceNorm2d(out_chans),
            torch.nn.LeakyReLU(negative_slope=0.2, inplace=True),
        )

    def forward(self, image: torch.Tensor) -> torch.Tensor:
        """Forward pass of :class:`TransposeConvBlock`.

        Parameters
        ----------
        image : torch.Tensor
            Input tensor of shape `(N, in_chans, H, W)`.

        Returns
        -------
        torch.Tensor
            Output tensor of shape `(N, out_chans, H*2, W*2)`.
        """
        return self.layers(image)



# coding=utf-8
__author__ = "Dimitris Karkalousos"

from typing import List, Tuple, Union

import torch
from omegaconf import DictConfig, OmegaConf
from pytorch_lightning import Trainer

from atommic.collections.multitask.rs.nn.base import BaseMRIReconstructionSegmentationModel
from atommic.collections.reconstruction.nn.unet_base.unet_block import Unet
from atommic.core.classes.common import typecheck

__all__ = ["RecSegUNet"]


class RecSegUNet(BaseMRIReconstructionSegmentationModel):
    """Implementation of the Reconstruction Segmentation method using UNets for both the reconstruction and
    segmentation as presented in [Sui2021]_.

    References
    ----------
    .. [Sui2021] Sui, B, Lv, J, Tong, X, Li, Y, Wang, C. Simultaneous image reconstruction and lesion segmentation in
        accelerated MRI using multitasking learning. Med Phys. 2021; 48: 7189– 7198. https://doi.org/10.1002/mp.15213

    """

    def __init__(self, cfg: DictConfig, trainer: Trainer = None):
        """Inits :class:`RecSegUNet`.

        Parameters
        ----------
        cfg : DictConfig
            Configuration object.
        trainer : Trainer, optional
            PyTorch Lightning trainer object. Default is ``None``.
        """
        super().__init__(cfg=cfg, trainer=trainer)

        cfg_dict = OmegaConf.to_container(cfg, resolve=True)

        self.input_channels = cfg_dict.get("input_channels", 2)
        if self.input_channels == 0:
            raise ValueError("Segmentation module input channels cannot be 0.")
        if self.input_channels > 2:
            raise ValueError(f"Segmentation module input channels must be either 1 or 2. Found: {self.input_channels}")

        reconstruction_module_output_channels = cfg_dict.get("reconstruction_module_output_channels", 1)

        self.reconstruction_module = Unet(
            in_chans=self.input_channels,
            out_chans=reconstruction_module_output_channels,
            chans=cfg_dict.get("reconstruction_module_channels", 64),
            num_pool_layers=cfg_dict.get("reconstruction_module_pooling_layers", 2),
            drop_prob=cfg_dict.get("reconstruction_module_dropout", 0.0),
        )

        self.segmentation_module = Unet(
            in_chans=reconstruction_module_output_channels,
            out_chans=cfg_dict.get("segmentation_module_output_channels", 1),
            chans=cfg_dict.get("segmentation_module_channels", 64),
            num_pool_layers=cfg_dict.get("segmentation_module_pooling_layers", 2),
            drop_prob=cfg_dict.get("segmentation_module_dropout", 0.0),
        )

        self.consecutive_slices = cfg_dict.get("consecutive_slices", 1)
        self.magnitude_input = cfg_dict.get("magnitude_input", True)
        self.normalize_segmentation_output = cfg_dict.get("normalize_segmentation_output", True)

    # pylint: disable=arguments-differ
    @typecheck()
    def forward(
        self,
        y: torch.Tensor,  # pylint: disable=unused-argument
        sensitivity_maps: torch.Tensor,  # pylint: disable=unused-argument
        mask: torch.Tensor,  # pylint: disable=unused-argument
        init_reconstruction_pred: torch.Tensor,
        target_reconstruction: torch.Tensor,  # pylint: disable=unused-argument
        hx: torch.Tensor = None,  # pylint: disable=unused-argument
        sigma: float = 1.0,  # pylint: disable=unused-argument
    ) -> Tuple[Union[List, torch.Tensor], torch.Tensor]:
        """Forward pass of :class:`RecSegUNet`.

        Parameters
        ----------
        y : torch.Tensor
            Subsampled k-space data. Shape [batch_size, n_coils, n_x, n_y, 2]
        sensitivity_maps : torch.Tensor
            Coil sensitivity maps. Shape [batch_size, n_coils, n_x, n_y, 2]
        mask : torch.Tensor
            Subsampling mask. Shape [1, 1, n_x, n_y, 1]
        init_reconstruction_pred : torch.Tensor
            Initial reconstruction prediction. Shape [batch_size, n_x, n_y, 2]
        target_reconstruction : torch.Tensor
            Target reconstruction. Shape [batch_size, n_x, n_y, 2]
        hx : torch.Tensor, optional
            Initial hidden state for the RNN. Default is ``None``.
        sigma : float, optional
            Standard deviation of the noise. Default is ``1.0``.

        Returns
        -------
        Tuple[Union[List, torch.Tensor], torch.Tensor]
            Tuple containing the predicted reconstruction and segmentation.
        """
        if self.consecutive_slices > 1:
            batch, slices = init_reconstruction_pred.shape[:2]
            init_reconstruction_pred = init_reconstruction_pred.reshape(
                init_reconstruction_pred.shape[0] * init_reconstruction_pred.shape[1],
                *init_reconstruction_pred.shape[2:],
            )

        if init_reconstruction_pred.shape[-1] == 2:
            if self.input_channels == 1:
                init_reconstruction_pred = torch.view_as_complex(init_reconstruction_pred).unsqueeze(1)
                if self.magnitude_input:
                    init_reconstruction_pred = torch.abs(init_reconstruction_pred)
            elif self.input_channels == 2:
                if self.magnitude_input:
                    raise ValueError("Magnitude input is not supported for 2-channel input.")
                init_reconstruction_pred = init_reconstruction_pred.permute(0, 3, 1, 2)
            else:
                raise ValueError(f"The input channels must be either 1 or 2. Found: {self.input_channels}")
        else:
            if init_reconstruction_pred.dim() == 3:
                init_reconstruction_pred = init_reconstruction_pred.unsqueeze(1)

        pred_reconstruction = self.reconstruction_module(torch.abs(init_reconstruction_pred))

        if self.magnitude_input:
            pred_reconstruction = torch.abs(pred_reconstruction)

        pred_segmentation = self.segmentation_module(pred_reconstruction)

        if self.normalize_segmentation_output:
            pred_segmentation = (pred_segmentation - pred_segmentation.min()) / (
                pred_segmentation.max() - pred_segmentation.min()
            )

        pred_segmentation = torch.abs(pred_segmentation)

        pred_reconstruction = pred_reconstruction.squeeze(1)

        if self.consecutive_slices > 1:
            pred_reconstruction = pred_reconstruction.view([batch, slices, *pred_reconstruction.shape[1:]])
            pred_segmentation = pred_segmentation.view([batch, slices, *pred_segmentation.shape[1:]])

        return pred_reconstruction, pred_segmentation



# coding=utf-8
__author__ = "Dimitris Karkalousos"


import torch
from omegaconf import DictConfig, OmegaConf
from pytorch_lightning import Trainer
from torch import nn

from atommic.collections.common.parts.utils import check_stacked_complex
from atommic.collections.reconstruction.nn.base import BaseMRIReconstructionModel
from atommic.collections.reconstruction.nn.modl_base.modl_block import ConjugateGradient, ResidualNetwork
from atommic.core.classes.common import typecheck

__all__ = ["MoDL"]


class MoDL(BaseMRIReconstructionModel):
    """Implementation of the MoDL: Model Based Deep Learning Architecture for Inverse Problems.

    Adjusted to optionally perform a data consistency step (Conjugate Gradient), as presented in [Aggarwal2018]_,
    [Yaman2020]_. If dc is set to False, the network will perform a simple residual learning step.

    References
    ----------
    .. [Aggarwal2018] MoDL: Model Based Deep Learning Architecture for Inverse Problems by H.K. Aggarwal, M.P Mani, and
        Mathews Jacob in IEEE Transactions on Medical Imaging, 2018

    .. [Yaman2020] Yaman, B, Hosseini, SAH, Moeller, S, Ellermann, J, Uğurbil, K, Akçakaya, M. Self-supervised
        learning of physics-guided reconstruction neural networks without fully sampled reference data. Magn Reson
        Med. 2020; 84: 3172– 3191. https://doi.org/10.1002/mrm.28378

    """

    def __init__(self, cfg: DictConfig, trainer: Trainer = None):
        """Inits :class:`MoDL`.

        Parameters
        ----------
        cfg : DictConfig
            Configuration.
        trainer : Trainer, optional
            PyTorch Lightning trainer. Default is ``None``.
        """
        super().__init__(cfg=cfg, trainer=trainer)

        cfg_dict = OmegaConf.to_container(cfg, resolve=True)

        self.unrolled_iterations = cfg_dict.get("unrolled_iterations", 10)
        self.reconstruction_module = ResidualNetwork(
            nb_res_blocks=cfg_dict.get("residual_blocks", 15),
            channels=cfg_dict.get("channels", 64),
            regularization_factor=cfg_dict.get("regularization_factor", 0.1),
        )
        self.dc = cfg_dict.get("conjugate_gradient_dc", False)
        if self.dc:
            self.mu = nn.Parameter(torch.Tensor([cfg_dict.get("penalization_weight")]), requires_grad=True)
            self.dc_block = ConjugateGradient(
                cfg_dict.get("conjugate_gradient_iterations", 10),
                self.mu,
                self.fft_centered,
                self.fft_normalization,
                self.spatial_dims,
                self.coil_dim,
                self.coil_combination_method,
            )

    # pylint: disable=arguments-differ
    @typecheck()
    def forward(
        self,
        y: torch.Tensor,  # pylint: disable=unused-argument
        sensitivity_maps: torch.Tensor,
        mask: torch.Tensor,
        initial_prediction: torch.Tensor,
        sigma: float = 1.0,  # pylint: disable=unused-argument
    ) -> torch.Tensor:
        """Forward pass of :class:`MoDL`.

        Parameters
        ----------
        y : torch.Tensor
            Subsampled k-space data. Shape [batch_size, n_coils, n_x, n_y, 2]
        sensitivity_maps : torch.Tensor
            Coil sensitivity maps. Shape [batch_size, n_coils, n_x, n_y, 2]
        mask : torch.Tensor
            Subsampling mask. Shape [1, 1, n_x, n_y, 1]
        initial_prediction : torch.Tensor
            Initial prediction. Shape [batch_size, n_x, n_y, 2]
        sigma : float
            Noise level. Default is ``1.0``.

        Returns
        -------
        torch.Tensor
            Prediction of the final cascade. Shape [batch_size, n_x, n_y]
        """
        x = initial_prediction.clone()
        for _ in range(self.unrolled_iterations):
            x = self.reconstruction_module(x.permute(0, 3, 1, 2)).permute(0, 2, 3, 1)
            if self.dc:
                x = self.dc_block(initial_prediction + self.mu * x, sensitivity_maps, mask)
        return check_stacked_complex(x)



# coding=utf-8
__author__ = "Dimitris Karkalousos"


import torch
from omegaconf import DictConfig, OmegaConf
from pytorch_lightning import Trainer

from atommic.collections.common.parts.fft import fft2, ifft2
from atommic.collections.common.parts.utils import complex_conj, complex_mul
from atommic.collections.reconstruction.nn.base import BaseMRIReconstructionModel
from atommic.collections.reconstruction.nn.ccnn_base.ccnn_block import Conv2d
from atommic.collections.reconstruction.nn.didn_base.didn_block import DIDN
from atommic.collections.reconstruction.nn.mwcnn_base.mwcnn_block import MWCNN
from atommic.collections.reconstruction.nn.primaldualnet_base.primaldualnet_block import DualNet, PrimalNet
from atommic.collections.reconstruction.nn.unet_base.unet_block import NormUnet
from atommic.core.classes.common import typecheck

__all__ = ["LPDNet"]


class LPDNet(BaseMRIReconstructionModel):
    """Implementation of the Learned Primal Dual network, inspired by [Adler2018]_.

    References
    ----------
    .. [Adler2018] Adler, Jonas, and Ozan Öktem. “Learned Primal-Dual Reconstruction.” IEEE Transactions on Medical
        Imaging, vol. 37, no. 6, June 2018, pp. 1322–32. arXiv.org, https://doi.org/10.1109/TMI.2018.2799231.

    """

    def __init__(self, cfg: DictConfig, trainer: Trainer = None):
        """Inits :class:`LPDNet`.

        Parameters
        ----------
        cfg : DictConfig
            Configuration.
        trainer : Trainer, optional
            PyTorch Lightning trainer. Default is ``None``.
        """
        super().__init__(cfg=cfg, trainer=trainer)

        cfg_dict = OmegaConf.to_container(cfg, resolve=True)

        self.dimensionality = cfg_dict.get("dimensionality", 2)
        self.consecutive_slices = cfg_dict.get("consecutive_slices", 1)

        self.num_iter = cfg_dict.get("num_iter")
        self.num_primal = cfg_dict.get("num_primal")
        self.num_dual = cfg_dict.get("num_dual")

        primal_model_architecture = cfg_dict.get("primal_model_architecture")

        if primal_model_architecture == "MWCNN":
            primal_model = torch.nn.Sequential(
                *[
                    MWCNN(
                        input_channels=cfg_dict.get("primal_in_channels") * (self.num_primal + 1),
                        first_conv_hidden_channels=cfg_dict.get("primal_mwcnn_hidden_channels"),
                        num_scales=cfg_dict.get("primal_mwcnn_num_scales"),
                        bias=cfg_dict.get("primal_mwcnn_bias"),
                        batchnorm=cfg_dict.get("primal_mwcnn_batchnorm"),
                    ),
                    torch.nn.Conv2d(
                        cfg_dict.get("primal_out_channels") * (self.num_primal + 1),
                        cfg_dict.get("primal_out_channels") * self.num_primal,
                        kernel_size=1,
                    ),
                ]
            )
        elif primal_model_architecture in ["UNET", "NORMUNET"]:
            primal_model = NormUnet(
                cfg_dict.get("primal_unet_num_filters"),
                cfg_dict.get("primal_unet_num_pool_layers"),
                in_chans=cfg_dict.get("primal_in_channels") * (self.num_primal + 1),
                out_chans=cfg_dict.get("primal_out_channels") * self.num_primal,
                drop_prob=cfg_dict.get("primal_unet_dropout_probability"),
                padding_size=cfg_dict.get("primal_unet_padding_size"),
                normalize=cfg_dict.get("primal_unet_normalize"),
            )
        else:
            raise NotImplementedError(
                "LPDNet is currently implemented for primal_model_architecture == 'CONV' or 'UNet'."
                f"Got primal_model_architecture == {primal_model_architecture}."
            )

        dual_model_architecture = cfg_dict.get("dual_model_architecture")

        if dual_model_architecture == "CONV":
            dual_model = Conv2d(
                in_channels=cfg_dict.get("dual_in_channels") * (self.num_dual + 2),
                out_channels=cfg_dict.get("dual_out_channels") * self.num_dual,
                hidden_channels=cfg_dict.get("kspace_conv_hidden_channels"),
                n_convs=cfg_dict.get("kspace_conv_n_convs"),
                batchnorm=cfg_dict.get("kspace_conv_batchnorm"),
            )
        elif dual_model_architecture == "DIDN":
            dual_model = DIDN(
                in_channels=cfg_dict.get("dual_in_channels") * (self.num_dual + 2),
                out_channels=cfg_dict.get("dual_out_channels") * self.num_dual,
                hidden_channels=cfg_dict.get("kspace_didn_hidden_channels"),
                num_dubs=cfg_dict.get("kspace_didn_num_dubs"),
                num_convs_recon=cfg_dict.get("kspace_didn_num_convs_recon"),
            )
        elif dual_model_architecture in ["UNET", "NORMUNET"]:
            dual_model = NormUnet(
                cfg_dict.get("dual_unet_num_filters"),
                cfg_dict.get("dual_unet_num_pool_layers"),
                in_chans=cfg_dict.get("dual_in_channels") * (self.num_dual + 2),
                out_chans=cfg_dict.get("dual_out_channels") * self.num_dual,
                drop_prob=cfg_dict.get("dual_unet_dropout_probability"),
                padding_size=cfg_dict.get("dual_unet_padding_size"),
                normalize=cfg_dict.get("dual_unet_normalize"),
            )
        else:
            raise NotImplementedError(
                "LPDNet is currently implemented for dual_model_architecture == 'CONV' or 'DIDN' or 'UNet'."
                f"Got dual_model_architecture == {dual_model_architecture}."
            )

        self.primal_net = torch.nn.ModuleList(
            [PrimalNet(self.num_primal, primal_architecture=primal_model) for _ in range(self.num_iter)]
        )
        self.dual_net = torch.nn.ModuleList(
            [DualNet(self.num_dual, dual_architecture=dual_model) for _ in range(self.num_iter)]
        )

    # pylint: disable=arguments-differ
    @typecheck()
    def forward(
        self,
        y: torch.Tensor,
        sensitivity_maps: torch.Tensor,
        mask: torch.Tensor,
        initial_prediction: torch.Tensor,
        sigma: float = 1.0,  # pylint: disable=unused-argument
    ) -> torch.Tensor:
        """Forward pass of :class:`LPDNet`.

        Parameters
        ----------
        y : torch.Tensor
            Subsampled k-space data. Shape [batch_size, n_coils, n_x, n_y, 2]
        sensitivity_maps : torch.Tensor
            Coil sensitivity maps. Shape [batch_size, n_coils, n_x, n_y, 2]
        mask : torch.Tensor
            Subsampling mask. Shape [1, 1, n_x, n_y, 1]
        initial_prediction : torch.Tensor
            Initial prediction. Shape [batch_size, n_x, n_y, 2]
        sigma : float
            Noise level. Default is ``1.0``.

        Returns
        -------
        torch.Tensor
            Prediction of the final cascade. Shape [batch_size, n_x, n_y]
        """
        dual_buffer = torch.cat([y] * self.num_dual, -1).to(y.device)
        primal_buffer = torch.cat([initial_prediction] * self.num_primal, -1).to(y.device)

        for idx in range(self.num_iter):
            # Dual
            f_2 = primal_buffer[..., 2:4].clone()
            f_2 = torch.where(
                mask == 0,
                torch.tensor([0.0], dtype=f_2.dtype).to(f_2.device),
                fft2(
                    complex_mul(f_2.unsqueeze(self.coil_dim), sensitivity_maps),
                    centered=self.fft_centered,
                    normalization=self.fft_normalization,
                    spatial_dims=self.spatial_dims,
                ).type(f_2.type()),
            )
            dual_buffer = self.dual_net[idx](dual_buffer, f_2, y)

            # Primal
            h_1 = dual_buffer[..., 0:2].clone()
            # needed for python3.9
            h_1 = torch.view_as_real(h_1[..., 0] + 1j * h_1[..., 1])
            h_1 = complex_mul(
                ifft2(
                    torch.where(mask == 0, torch.tensor([0.0], dtype=h_1.dtype).to(h_1.device), h_1),
                    centered=self.fft_centered,
                    normalization=self.fft_normalization,
                    spatial_dims=self.spatial_dims,
                ),
                complex_conj(sensitivity_maps),
            ).sum(self.coil_dim)
            primal_buffer = self.primal_net[idx](primal_buffer, h_1)

        primal_buffer = primal_buffer[..., 0:2]

        return torch.view_as_real(primal_buffer[..., 0] + 1j * primal_buffer[..., 1])



# coding=utf-8
__author__ = "Dimitris Karkalousos"

from typing import Sequence

import torch
from torch import nn

from atommic.collections.common.parts.fft import fft2, ifft2
from atommic.collections.common.parts.utils import coil_combination_method


class ResidualNetwork(nn.Module):
    """Residual Network, as presented in [Aggarwal2018]_, [Yaman2020]_.

    References
    ----------
    [Aggarwal2018] MoDL: Model Based Deep Learning Architecture for Inverse Problems by H.K. Aggarwal, M.P Mani, and
        Mathews Jacob in IEEE Transactions on Medical Imaging, 2018

    [Yaman2020] Yaman, B, Hosseini, SAH, Moeller, S, Ellermann, J, Uğurbil, K, Akçakaya, M. Self-supervised learning of
        physics-guided reconstruction neural networks without fully sampled reference data. Magn Reson Med. 2020; 84:
        3172– 3191. https://doi.org/10.1002/mrm.28378
    """

    def __init__(self, nb_res_blocks: int = 15, channels: int = 64, regularization_factor: float = 0.1):
        """Inits :class:`ResidualNetwork`.

        Parameters
        ----------
        nb_res_blocks : int, optional
            Number of residual blocks. Default is ``15``.
        channels : int, optional
            Number of channels. Default is ``64``.
        regularization_factor : float, optional
            Regularization factor. Default is ``0.1``.
        """
        super().__init__()
        self.relu = nn.ReLU(inplace=True)
        self.conv1 = nn.Conv2d(2, channels, kernel_size=3, stride=1, padding="same", bias=False)
        self.layers1 = nn.ModuleList()
        self.layers2 = nn.ModuleList()
        for _ in range(1, nb_res_blocks + 1):
            self.layers1.append(nn.Conv2d(channels, channels, kernel_size=3, stride=1, padding="same", bias=False))
            self.layers2.append(nn.Conv2d(channels, channels, kernel_size=3, stride=1, padding="same", bias=False))
        self.last_layer = nn.Conv2d(channels, channels, kernel_size=3, stride=1, padding="same", bias=False)
        self.final_layer = nn.Conv2d(channels, 2, kernel_size=3, stride=1, padding="same", bias=False)
        self.scaling = torch.tensor([regularization_factor]).type(torch.float32)
        self.__weights_initialization__()

    def __weights_initialization__(self):
        """Initializes the weights of the network."""
        for m in self.modules():
            if isinstance(m, nn.Conv2d):
                nn.init.xavier_normal_(m.weight)
                if m.bias is not None:
                    nn.init.zeros_(m.bias)

    def forward(self, x: torch.Tensor) -> torch.Tensor:
        """Forward pass of :class:`ResidualNetwork`."""
        out = self.conv1(x)
        x = out
        for i in range(len(self.layers1)):  # pylint: disable=consider-using-enumerate
            x = self.scaling.to(x.device) * self.layers2[i](self.relu(self.layers1[i](x))) + x
        x = self.last_layer(x)
        x = x + out
        return self.final_layer(x)


class ConjugateGradient(nn.Module):
    """Conjugate Gradient algorithm for solving the linear system of equations, as presented in [Aggarwal2018]_,
    [Yaman2020]_.

    References
    ----------
    [Aggarwal2018] MoDL: Model Based Deep Learning Architecture for Inverse Problems by H.K. Aggarwal, M.P Mani, and
        Mathews Jacob in IEEE Transactions on Medical Imaging, 2018

    [Yaman2020] Yaman, B, Hosseini, SAH, Moeller, S, Ellermann, J, Uğurbil, K, Akçakaya, M. Self-supervised learning of
        physics-guided reconstruction neural networks without fully sampled reference data. Magn Reson Med. 2020; 84:
        3172– 3191. https://doi.org/10.1002/mrm.28378
    """

    def __init__(
        self,
        CG_Iter: int = 10,
        mu: nn.Parameter = nn.Parameter(torch.tensor([0.05]).type(torch.float32)),
        fft_centered: bool = False,
        fft_normalization: str = "ortho",
        spatial_dims: Sequence[int] = (-2, -1),
        coil_dim: int = 1,
        coil_combination_method: str = "SENSE",
    ):
        """Inits :class:`ConjugateGradient`.

        Parameters
        ----------
        CG_Iter : int, optional
            Number of CG iterations. Default is ``10``.
        mu : torch.nn.Parameter, optional
            Regularization parameter. Default is ``0.05``.
        fft_centered : bool, optional
            Whether to center the FFT. Default is ``False``.
        fft_normalization : str, optional
            Normalization type of the FFT. Default is ``"ortho"``.
        spatial_dims : Sequence[int], optional
            Spatial dimensions of the input. Default is ``(-2, -1)``.
        coil_dim : int, optional
            Coil dimension. Default is ``1``.
        """
        super().__init__()
        self.CG_Iter = CG_Iter
        self.mu = mu
        self.fft_centered = fft_centered
        self.fft_normalization = fft_normalization
        self.spatial_dims = spatial_dims
        self.coil_dim = coil_dim
        self.coil_combination_method = coil_combination_method

    def EhE_Op(
        self, prediction: torch.Tensor, sens_maps: torch.Tensor, mask: torch.Tensor  # pylint: disable=unused-argument
    ) -> torch.Tensor:
        """This function calculates the product of the operator EhE with a given vector.

        Parameters
        ----------
        prediction : torch.Tensor
            The input vector.
        sens_maps : torch.Tensor
            The sensitivity maps.
        mask : torch.Tensor
            The undersampling mask.

        Returns
        -------
        torch.Tensor
            Data consistency term.
        """
        masked_kspace = fft2(
            prediction.unsqueeze(self.coil_dim) * sens_maps,
            self.fft_centered,
            self.fft_normalization,
            self.spatial_dims,
        )
        masked_kspace = torch.view_as_real(masked_kspace[..., 0] + 1j * masked_kspace[..., 1])
        image_space = ifft2(masked_kspace, self.fft_centered, self.fft_normalization, self.spatial_dims)
        pred = coil_combination_method(
            image_space, torch.view_as_real(sens_maps), self.coil_combination_method, self.coil_dim
        )
        pred = torch.view_as_real(pred[..., 0] + 1j * pred[..., 1])
        return torch.view_as_complex(pred) + self.mu * prediction

    def proximal_gradient(
        self,
        rsold: torch.Tensor,
        x: torch.Tensor,
        r: torch.Tensor,
        p: torch.Tensor,
        sens_maps: torch.Tensor,
        mask: torch.Tensor,
    ) -> torch.Tensor:
        """Proximal gradient descent.

        Parameters
        ----------
        rsold : torch.Tensor
            The residual.
        x : torch.Tensor
            The current estimate.
        r : torch.Tensor
            The current residual.
        p : torch.Tensor
            The current search direction.
        sens_maps : torch.Tensor
            The sensitivity maps.
        mask : torch.Tensor
            The undersampling mask.

        Returns
        -------
        torch.Tensor
            The new residual, the new estimate, the new residual, and the new search direction.
        """
        Ap = self.EhE_Op(p, sens_maps, mask)
        alpha = rsold / torch.sum(torch.conj(p) * Ap)
        alpha = alpha + 0j
        x = x + alpha * p
        r = r - alpha * Ap
        rsnew = torch.sum(torch.conj(r) * r)
        beta = rsnew / rsold
        beta = beta + 0j
        p = r + beta * p
        return rsnew, x, r, p

    def forward(self, rhs: torch.Tensor, sens_maps: torch.Tensor, mask: torch.Tensor) -> torch.Tensor:
        """Forward pass of :class:`ConjugateGradient`.

        Parameters
        ----------
        rhs : torch.Tensor
            The right-hand side of the linear system.
        sens_maps : torch.Tensor
            The sensitivity maps.
        mask : torch.Tensor
            The undersampling mask.

        Returns
        -------
        torch.Tensor
            The solution of the linear system.
        """
        rhs = torch.view_as_complex(rhs)
        sens_maps = torch.view_as_complex(sens_maps)
        x = torch.zeros_like(rhs)
        i, r, p = 0, rhs, rhs
        rsold = torch.sum(torch.conj(r) * r)
        while i < self.CG_Iter:
            rsold, x, r, p = self.proximal_gradient(rsold, x, r, p, sens_maps, mask)
            i = i + 1
        return torch.view_as_real(x)



# coding=utf-8
__author__ = "Dimitris Karkalousos"

# Taken and adapted from: https://github.com/NVIDIA/NeMo/blob/main/nemo/core/classes/exportable.py

from abc import ABC
from typing import List, Union

import torch
from pytorch_lightning.core.module import _jit_is_scripting
from torch.onnx import TrainingMode

from atommic.core.classes.common import typecheck
from atommic.core.utils.neural_type_utils import get_dynamic_axes, get_io_names
from atommic.utils import logging
from atommic.utils.export_utils import (
    ExportFormat,
    augment_filename,
    get_export_format,
    parse_input_example,
    replace_for_export,
    verify_runtime,
    verify_torchscript,
    wrap_forward_method,
)

__all__ = ["ExportFormat", "Exportable"]


class Exportable(ABC):
    """This Interface should be implemented by particular classes derived from atommic.core.ModelPT. It gives these
    entities ability to be exported for deployment to formats such as ONNX.
    """

    @property
    def input_module(self):
        """Implement this method to return the input module"""
        return self

    @property
    def output_module(self):
        """Implement this method to return the output module."""
        return self

    def export(
        self,
        output: str,
        input_example=None,
        verbose=False,
        do_constant_folding=True,
        onnx_opset_version=None,
        check_trace: Union[bool, List[torch.Tensor]] = False,
        dynamic_axes=None,
        check_tolerance=0.01,
        export_modules_as_functions: bool = False,
        keep_initializers_as_inputs=None,
    ):
        """Export the module to a file.

        Parameters
        ----------
        output : str
            The output file path.
        input_example : dict
            A dictionary of input names and values.
        verbose : bool
            If True, print out the export process.
        do_constant_folding : bool
            If True, do constant folding.
        onnx_opset_version : int
            The ONNX opset version to use.
        check_trace : bool or list of torch.Tensor
            If True, check the trace of the exported model.
        dynamic_axes : dict
            A dictionary of input names and dynamic axes.
        check_tolerance : float
            The tolerance for the check_trace.
        export_modules_as_functions : bool
            If True, export modules as functions.
        keep_initializers_as_inputs : bool
            If True, keep initializers as inputs.
        """
        all_out = []
        all_descr = []
        for subnet_name in self.list_export_subnets():
            model = self.get_export_subnet(subnet_name)
            out_name = augment_filename(output, subnet_name)
            out, descr, out_example = model._export(  # pylint: disable=protected-access
                out_name,
                input_example=input_example,
                verbose=verbose,
                do_constant_folding=do_constant_folding,
                onnx_opset_version=onnx_opset_version,
                check_trace=check_trace,
                dynamic_axes=dynamic_axes,
                check_tolerance=check_tolerance,
                export_modules_as_functions=export_modules_as_functions,
                keep_initializers_as_inputs=keep_initializers_as_inputs,
            )
            # Propagate input example (default scenario, may need to be overriden)
            if input_example is not None:
                input_example = out_example
            all_out.append(out)
            all_descr.append(descr)
            logging.info(f"Successfully exported {model.__class__.__name__} to {out_name}")
        return all_out, all_descr

    def _export(
        self,
        output: str,
        input_example=None,
        verbose=False,
        do_constant_folding=True,
        onnx_opset_version=None,
        training=TrainingMode.EVAL,  # pylint: disable=unused-argument
        check_trace: Union[bool, List[torch.Tensor]] = False,
        dynamic_axes=None,
        check_tolerance=0.01,
        export_modules_as_functions: bool = False,
        keep_initializers_as_inputs=None,
    ):
        """Helper to export the module to a file.

        Parameters
        ----------
        output : str
            The output file path.
        input_example : dict
            A dictionary of input names and values.
        verbose : bool
            If True, print out the export process.
        do_constant_folding : bool
            If True, do constant folding.
        onnx_opset_version : int
            The ONNX opset version to use.
        training : TrainingMode
            Training mode for the export.
        check_trace : bool or list of torch.Tensor
            If True, check the trace of the exported model.
        dynamic_axes : dict
            A dictionary of input names and dynamic axes.
        check_tolerance : float
            The tolerance for the check_trace.
        export_modules_as_functions : bool
            If True, export modules as functions.
        keep_initializers_as_inputs : bool
            If True, keep initializers as inputs.
        """
        my_args = locals().copy()
        my_args.pop("self")

        self.eval()  # type: ignore
        for param in self.parameters():  # type: ignore
            param.requires_grad = False

        exportables = [m for m in self.modules() if isinstance(m, Exportable)]  # type: ignore
        qual_name = f"{self.__module__}.{self.__class__.__qualname__}"
        exp_format = get_export_format(output)
        output_descr = f"{qual_name} exported to {exp_format}"

        # Pytorch's default opset version is too low, using reasonable latest one
        if onnx_opset_version is None:
            onnx_opset_version = 16

        try:
            # Disable typechecks
            typecheck.set_typecheck_enabled(enabled=False)

            # Allow user to completely override forward method to export
            forward_method, old_forward_method = wrap_forward_method(self)

            with torch.inference_mode(), torch.no_grad(), torch.jit.optimized_execution(True), _jit_is_scripting():
                if input_example is None:
                    input_example = self.input_module.input_example()

                # Remove i/o examples from args we propagate to enclosed Exportables
                my_args.pop("output")
                my_args.pop("input_example")

                # Run (possibly overridden) prepare methods before calling forward()
                for ex in exportables:
                    ex._prepare_for_export(**my_args, noreplace=True)  # pylint: disable=protected-access
                self._prepare_for_export(output=output, input_example=input_example, **my_args)

                input_list, input_dict = parse_input_example(input_example)
                input_names = self.input_names
                output_names = self.output_names
                output_example = tuple(self.forward(*input_list, **input_dict))  # type: ignore

                if check_trace:
                    if isinstance(check_trace, bool):
                        check_trace_input = [input_example]
                    else:
                        check_trace_input = check_trace
                jitted_model = self
                if exp_format == ExportFormat.TORCHSCRIPT:
                    jitted_model = torch.jit.trace_module(
                        self,
                        {"forward": tuple(input_list) + tuple(input_dict.values())},
                        strict=True,
                        check_trace=check_trace,
                        check_tolerance=check_tolerance,
                    )
                    jitted_model = torch.jit.freeze(jitted_model)
                    if verbose:
                        logging.info(f"JIT code:\n{jitted_model.code}")  # type: ignore
                    jitted_model.save(output)  # type: ignore
                    jitted_model = torch.jit.load(output)

                    if check_trace:
                        verify_torchscript(jitted_model, output, check_trace_input, check_tolerance)
                elif exp_format == ExportFormat.ONNX:
                    # dynamic axis is a mapping from input/output_name => list of "dynamic" indices
                    if dynamic_axes is None:
                        dynamic_axes = get_dynamic_axes(self.input_module.input_types_for_export, input_names)
                        dynamic_axes.update(get_dynamic_axes(self.output_module.output_types_for_export, output_names))
                    torch.onnx.export(
                        jitted_model,
                        input_example,
                        output,
                        input_names=input_names,
                        output_names=output_names,
                        verbose=verbose,
                        do_constant_folding=do_constant_folding,
                        dynamic_axes=dynamic_axes,
                        opset_version=onnx_opset_version,
                        keep_initializers_as_inputs=keep_initializers_as_inputs,
                        export_modules_as_functions=export_modules_as_functions,
                    )

                    if check_trace:
                        verify_runtime(self, output, check_trace_input, input_names, check_tolerance=check_tolerance)
                else:
                    raise ValueError(f"Encountered unknown export format {exp_format}.")
        finally:
            typecheck.set_typecheck_enabled(enabled=True)
            if forward_method:  # pylint: disable=used-before-assignment
                type(self).forward = old_forward_method  # type: ignore  # pylint: disable=used-before-assignment
            self._export_teardown()
        return (output, output_descr, output_example)

    @property
    def disabled_deployment_input_names(self):
        """Implement this method to return a set of input names disabled for export"""
        return set()

    @property
    def disabled_deployment_output_names(self):
        """Implement this method to return a set of output names disabled for export"""
        return set()

    @property
    def supported_export_formats(self):
        """Implement this method to return a set of export formats supported. Default is all types."""
        return {ExportFormat.ONNX, ExportFormat.TORCHSCRIPT}

    def _prepare_for_export(self, **kwargs):
        """Override this method to prepare module for export. This is in-place operation. Base version does common
        necessary module replacements (Apex etc)
        """
        if "noreplace" not in kwargs:
            replace_for_export(self)

    def _export_teardown(self):
        """Override this method for any teardown code after export."""

    @property
    def input_names(self):
        """Implement this method to return a list of input names"""
        return get_io_names(self.input_module.input_types_for_export, self.disabled_deployment_input_names)

    @property
    def output_names(self):
        """Override this method to return a set of output names disabled for export"""
        return get_io_names(self.output_module.output_types_for_export, self.disabled_deployment_output_names)

    @property
    def input_types_for_export(self):
        """Implement this method to return a list of input types"""
        return self.input_types

    @property
    def output_types_for_export(self):
        """Implement this method to return a list of output types"""
        return self.output_types

    def get_export_subnet(self, subnet=None):
        """Returns Exportable subnet model/module to export"""
        return self if subnet is None or subnet == "self" else getattr(self, subnet)

    @staticmethod
    def list_export_subnets():
        """Returns default set of subnet names exported for this model. First goes the one receiving input
        (input_example).
        """
        return ["self"]

    def get_export_config(self):
        """Returns export_config dictionary."""
        return getattr(self, 'export_config', {})

    def set_export_config(self, args):
        """Sets/updates export_config dictionary."""
        ex_config = self.get_export_config()
        ex_config.update(args)
        self.export_config = ex_config



# coding=utf-8
__author__ = "Dimitris Karkalousos"

# Taken and adapted from: https://github.com/NVIDIA/NeMo/blob/main/nemo/utils/formatters/base.py

import logging

from atommic.utils.formaters.colors import Fore as ForegroundColors
from atommic.utils.formaters.utils import check_color_support, to_unicode

__all__ = ["BaseATOMMICFormatter", "DebugATOMMICFormatter"]


class BaseFormatter(logging.Formatter):
    """Base class for all formatters used in Tornado. Key features of this formatter are:
    * Color support when logging to a terminal that supports it.
    * Timestamps on every log line.
    * Robust against str/bytes encoding problems.
    """

    DEFAULT_FORMAT = "%(color)s[%(levelname)1.1s %(asctime)s %(module)s:%(lineno)d]%(end_color)s %(message)s"

    DEFAULT_DATE_FORMAT = "%Y-%m-%d %H:%M:%S"

    DEFAULT_COLORS = {
        logging.DEBUG: ForegroundColors.CYAN,
        logging.INFO: ForegroundColors.GREEN,
        logging.WARNING: ForegroundColors.YELLOW,
        logging.ERROR: ForegroundColors.MAGENTA,
        logging.CRITICAL: ForegroundColors.RED,
    }

    def __init__(self, color=True, fmt=None, datefmt=None, colors=None):
        """Inits :class:`BaseFormatter`.

        Parameters
        ----------
        color : bool
            Enable color support. Default is ``True``.
        fmt : str
            Log message format. It will be applied to the attributes dict of log records. The text between
            ``%(color)s`` and ``%(end_color)s`` will be colored depending on the level if color support is on. Default
            is ``None``.
        datefmt : str
            Datetime format. Used for formatting ``(asctime)`` placeholder in ``prefix_fmt``. Default  is ``None``.
        colors : dict
            Dictionary mapping logging level to terminal color code. Default  is ``None``.
        """
        if fmt is None:
            fmt = self.DEFAULT_FORMAT

        if datefmt is None:
            datefmt = self.DEFAULT_DATE_FORMAT

        if colors is None:
            colors = self.DEFAULT_COLORS

        logging.Formatter.__init__(self, datefmt=datefmt)

        self._fmt = fmt
        self._colors = {}
        self._normal = ""

        if color and check_color_support():
            self._colors = colors
            self._normal = ForegroundColors.RESET

    def format(self, record):
        """Formats a record.

        Parameters
        ----------
        record : LogRecord
            Log record to be formatted.

        Returns
        -------
        str
            The formatted record as a string.
        """
        try:
            message = record.getMessage()
            if not isinstance(message, str):
                raise AssertionError
            # Encoding notes: The logging module prefers to work with character strings, but only enforces that log
            # messages are instances of asestring. In python 2, non-ascii bytestrings will make their way through the
            # logging framework until they blow up with an unhelpful decoding error (with this formatter it happens
            # when we attach the prefix, but there are other opportunities for exceptions further along in the
            # framework).
            #
            # If a byte string makes it this far, convert it to unicode to ensure it will make it out to the logs. Use
            # repr() as a fallback to ensure that all byte strings can be converted successfully, but don't do it by
            # default, so we don't add extra quotes to ascii bytestrings. This is a bit of a hacky place to do this,
            # but it's worth it since the encoding errors that would otherwise result are so useless (and tornado is
            # fond of using utf8-encoded byte strings wherever possible).
            record.message = to_unicode(message)

        except Exception as e:
            record.message = f"Bad message (%{e}): %{record.__dict__}"

        record.asctime = self.formatTime(record, self.datefmt)

        if record.levelno in self._colors:
            record.color = self._colors[record.levelno]
            record.end_color = self._normal
        else:
            record.color = record.end_color = ""

        formatted = self._fmt % record.__dict__

        if record.exc_info and not record.exc_text:
            record.exc_text = self.formatException(record.exc_info)

        if record.exc_text:
            # exc_text contains multiple lines.  We need to _safe_unicode
            # each line separately so that non-utf8 bytes don't cause all the newlines to turn into '\n'.
            lines = [formatted.rstrip()]
            lines.extend(to_unicode(ln) for ln in record.exc_text.split("\n"))

            formatted = "\n".join(lines)
        return formatted.replace("\n", "\n    ")


class BaseATOMMICFormatter(BaseFormatter):
    """Base formatter for atommic logs."""

    DEFAULT_FORMAT = "%(color)s[atommic %(levelname)1.1s %(asctime)s %(module)s:%(lineno)d]%(end_color)s %(message)s"


class DebugATOMMICFormatter(BaseFormatter):
    """Debug formatter for atommic logs."""

    DEFAULT_FORMAT = (
        "%(color)s[atommic %(levelname)1.1s %(asctime)s %(module)s:%(lineno)d rank:%(rank)s]%(end_color)s %(message)s"
    )
